
(use format-modular srfi-1)

(define *tests* '())

(define-record test group-name name proc expect equal?)

(define (register-test . args)
  (set! *tests* (cons (apply make-test args) *tests*)))

(define (run-test test)
  (when (getenv "TESTS_VERBOSE")
    (format (current-error-port) "Test~A: ~S..."
            (if (test-group-name test)
              (format #f " from group ~A" (test-group-name test))
              "")
            (test-name test)))
  (let* ((result ((test-proc test)))
         (expect ((test-expect test)))
         (pass ((test-equal? test) result expect)))
    (when (getenv "TESTS_VERBOSE")
      (format (current-error-port) " ~A~%" (if pass "PASS" "FAIL")))
    (unless pass
      (format (current-error-port) "Test~A failed: ~S~%Expected: ~S~%Received: ~S~%~%"
        (if (test-group-name test)
          (format #f " from group ~A" (test-group-name test))
          "")
        (test-name test)
        expect
        result))
    pass))

(define (run-tests)
  (when (getenv "TESTS_SHOW_GROUPS")
    (format (current-error-port) "Groups:~{ ~A~}~%"
            (delete-duplicates (map test-group-name (reverse *tests*)))))
  (when (and (getenv "TESTS")
             (positive?
               (count
                 (complement run-test)
                 (reverse
                   (filter
                     (let ((groups (map string->symbol
                                        (string-split
                                          (or (getenv "TESTS_GROUPS") "")
                                          " "))))
                       (if (null? groups)
                         identity
                         (compose (cut member <> groups)
                                  test-group-name)))
                     *tests*)))))
    (error "Unit tests failed")))
